/*
 *  Libjpeg interface routines.
 *  We support JPEG decompress interface only.
 */
#include "viewer.h"
#include "./jpeg/jpeglib.h"
#include "ff.h"
#include "tinylib.h"
#include "monitor.h"
#include "LCDpanel.h"
#include "grphcon.h"

#define	POOL_SIZE	(1024L*40L)       /* Hope this is big enough */

/*
 *  WJԂƎgpʂጸ邽߂ɁAJPEG̉摜TCY
 *  PIC_WIDTH X PIC_HEIGHT sNZɐ.
 */
#define	PIC_WIDTH	400
#define	PIC_HEIGHT	300

struct jpeg_error_mgr jerr;
struct jpeg_error_mgr *jpeg_jsp_error (struct jpeg_error_mgr * err);
void jpeg_fatfs_src (j_decompress_ptr cinfo, FIL * infile);

struct jpeg_decompress_struct dcinfo;

static FIL jpeg_file;
char mem_pool[POOL_SIZE] __attribute__ ((section ("stack")));

static char *pool_ptr;

void pool_init()
{
    pool_ptr = mem_pool;
}

void pool_stat()
{
    xprintf("%d used.\n", pool_ptr - mem_pool);
}

char *malloc(int size)
{
    char *mp;

    mp = pool_ptr;
    pool_ptr += size;
    if (pool_ptr - mem_pool > POOL_SIZE) {
        xprintf("No memory!\n");
        return NULL;
    }
    return mp;
}

void free(char *mp)
{
    /* DO NOTHING */
}

FIL *open_jpeg_file(char *fname)
{
    FRESULT result;

    result = f_open(&jpeg_file, fname, FA_READ);
    if (result != FR_OK) {
        return NULL;
    }
    return &jpeg_file;
}

void close_jpeg_file(FIL *fptr)
{
    f_close(fptr);
}

static int jmethod;

void select_jpeg_method(int flag)
{
    if (flag) {
	jmethod = JDCT_FLOAT;
    } else {
	jmethod = JDCT_ISLOW;
    }
}

/* 
 *  JPEG 摜WJ
 *    fname:  JPEG t@C
 *    orientation:  JPEG 摜̕\
 */
int expand_jpeg(char *fname, int orientation)
{
    int denom, scale;
    FIL *fptr;
    JSAMPARRAY buffer;
    int row_stride;
    int xoffset, yoffset, line;
    int hoffset, voffset;
    int picw, pich, maxw, maxh;
    int screenw, screenh;
    int dw;

xprintf("fname = %s\n", fname);
    if (orientation < 1 || orientation > 8)
	orientation = 1;
    pool_init();
    fptr = open_jpeg_file(fname);
    if (!fptr)  return 0;

    dcinfo.err = jpeg_jsp_error(&jerr);
    jpeg_create_decompress(&dcinfo);
    jpeg_fatfs_src(&dcinfo, fptr);

    jpeg_read_header(&dcinfo, TRUE);

    /*
     * screenw, screenh: LCDʂ̕A
     * picw, pich:       zXN[̃TCYBLCD̎ʃTCY傫
     * maxw, maxh:       JPEG摜̍őTCY. pcw/pich 8{.
     */
    if (orientation < 5) {	/* Landscape */
        screenw = LCD_WIDTH;
        screenh = LCD_HEIGHT;
	picw = PIC_WIDTH;
	pich = PIC_HEIGHT;
        maxw = PIC_WIDTH * 8;
        maxh = PIC_HEIGHT * 8;
    } else {			/* Portrait */
	screenw = LCD_HEIGHT;
	screenh = LCD_WIDTH;
	picw = PIC_HEIGHT;
	pich = PIC_WIDTH;
        maxw = PIC_HEIGHT * 8;
	maxh = PIC_WIDTH * 8;
    }

    /*
     * {\tgł́Amaxw * maxh 傫ȉ摜͈ȂBG[ƂB
     */
    if ((dcinfo.image_width > maxw) || (dcinfo.image_height > maxh)) {
        xprintf("image too big: %d x %d\n", dcinfo.image_width, dcinfo.image_height);
        /* Image is too big */
        jpeg_destroy_decompress(&dcinfo);
        close_jpeg_file(fptr);
        return 0;	/* G[ */
    }
    dcinfo.scale_denom = dcinfo.scale_num = 8;
    denom = 8;

    for (scale = 8; scale > 1; scale--) {
        if (dcinfo.image_width / denom * scale <= picw)
            break;
    }
    while (dcinfo.image_height/denom * scale > pich) {
        scale--;
    }

    if (scale > 0) {
        dcinfo.scale_num = scale;
        dcinfo.out_color_space = JCS_RGB;
        dcinfo.dct_method = jmethod;
        jpeg_start_decompress(&dcinfo);
xprintf("%d x %d --> %d x %d\n",
  dcinfo.image_width, dcinfo.image_height,
  dcinfo.output_width, dcinfo.output_height);

        xoffset = yoffset = 0;
	hoffset = voffset = 0;

        /* wiœhԂ */
	GRPH_Clear(C_BLACK);

        /* 摜Z^Oĕ\ */
        if (dcinfo.output_width <= screenw) {
                xoffset = (screenw - dcinfo.output_width) >> 1;
	} else {
		hoffset = (dcinfo.output_width - screenw) >> 1;
	}
	if (dcinfo.output_height <= screenh) {
                yoffset = (screenh - dcinfo.output_height) >> 1;
        } else {
		voffset = (dcinfo.output_height - screenh) >> 1;
	}
xprintf("hoffset = %d, voffset = %d\n", hoffset, voffset);

        row_stride = dcinfo.output_width * dcinfo.output_components;
        buffer = (*dcinfo.mem->alloc_sarray)((j_common_ptr)&dcinfo, JPOOL_IMAGE, row_stride, 1);
	dw = dcinfo.output_width;
	dw -= (hoffset << 1);
 
        while (dcinfo.output_scanline < dcinfo.output_height) {
	    line = dcinfo.output_scanline;
            jpeg_read_scanlines(&dcinfo, buffer, 1);
	    if (line >= voffset) {
		LCD_Draw_Line(orientation, buffer[0] + hoffset * 3,
		    xoffset, yoffset + line - voffset, dw);
	    }
        }
        pool_stat();

        jpeg_finish_decompress(&dcinfo);
    }
    jpeg_destroy_decompress(&dcinfo);
    close_jpeg_file(fptr);

    return 1;
}

/* 
 *  JPEG TlC摜̐LA\
 *    fname:  JPEGt@C
 *    thumb_offet:  JPEGt@CATlC摜̂ʒuւ̃ItZbg
 *    picno:        \ʒuԍ (0..11)
 *
 * Picno ̒lɉāALCDʏ̎̈ʒuɉ摜\
 *
 *   0: xoffset = 0,   yoffset = 0
 *   1: xoffset = 80,  yoffset = 0
 *   2: xoffset = 160, yoffset = 0
 *   3: xoffset = 240, yoffset = 0
 *   4: xoffset = 0,   yoffset = 60
 *   5: xoffset = 80,  yoffset = 60
 *   6: xoffset = 160, yoffset = 60
 *   7: xoffset = 240, yoffset = 60
 *   8: xoffset = 0,   yoffset = 120
 *   9: xoffset = 80,  yoffset = 120
 *  10: xoffset = 160, yoffset = 120
 *  11: xoffset = 240, yoffset = 120
 */
int expand_thumbnail(char *fname, long thumb_offset, int picno)
{
    int scale;
    int orientation;
    FIL *fptr;
    JSAMPARRAY buffer;
    int row_stride;
    int xoffset, yoffset, line;
    int dw, mod12;

#ifdef DEBUG
    xprintf("fname = %s\n", fname);
#endif
    pool_init();
    fptr = open_jpeg_file(fname);
    if (!fptr)  return 0;

    /* For thumbnail, seek to its start position */
    f_lseek(fptr, thumb_offset);
	
    dcinfo.err = jpeg_jsp_error(&jerr);
    jpeg_create_decompress(&dcinfo);
    jpeg_fatfs_src(&dcinfo, fptr);

    jpeg_read_header(&dcinfo, TRUE);

    /*
     * Thumbnail is always displayed in landscape mode.
     */
    orientation = 1;

    /* Verify picture size */
    if ((dcinfo.image_width != 160) || (dcinfo.image_height != 120)) {
        xprintf("Invalid thumbnail size: %d x %d\n", dcinfo.image_width, dcinfo.image_height);
        /* Image is too big */
        jpeg_destroy_decompress(&dcinfo);
        close_jpeg_file(fptr);
        return 0;
    }
    dcinfo.scale_denom = dcinfo.scale_num = 8;

    if (picno >= 0) {
	/*
	 * JPEG thumbnail picture is 160 x 120 dots.
	 * We'll scale it to half size (80 x 60)
         */
	scale = 4;	/* scale 4/8 */
    } else {
	scale = 8;	/* Use original 160 x 120 dots */
    }

    dcinfo.scale_num = scale;
    dcinfo.out_color_space = JCS_RGB;
    dcinfo.dct_method = jmethod;
    jpeg_calc_output_dimensions(&dcinfo);
#ifdef DEBUG
    xprintf("%d x %d --> %d x %d\n",
        dcinfo.image_width, dcinfo.image_height,
        dcinfo.output_width, dcinfo.output_height);
#endif

    xoffset = yoffset = 0;
    if (picno > 0) {
        mod12 = picno % 12;
        xoffset = 80 * (mod12 & 3);
        yoffset = 60 * (mod12 / 4);
    }
    jpeg_start_decompress(&dcinfo);

    row_stride = dcinfo.output_width * dcinfo.output_components;
    buffer = (*dcinfo.mem->alloc_sarray)((j_common_ptr)&dcinfo, JPOOL_IMAGE, row_stride, 1);
    dw = dcinfo.output_width;
 
    while (dcinfo.output_scanline < dcinfo.output_height) {
	line = dcinfo.output_scanline;
        jpeg_read_scanlines(&dcinfo, buffer, 1);
	LCD_Draw_Line(orientation, buffer[0], xoffset, yoffset + line, dw);
    }
#ifdef DEBUG
    pool_stat();
#endif

    jpeg_finish_decompress(&dcinfo);
    jpeg_destroy_decompress(&dcinfo);
    close_jpeg_file(fptr);

    return 1;
}
