/*------------------------------------------------------------------------*/
/* MMCv3/SDv1/SDv2 (SPI[h) 䃂W[                             */
/* C^[tF[X t^SH7262TvvO                      */
/*------------------------------------------------------------------------*/
/*
/  Copyright (C) 2010, ChaN, all right reserved.
/
/ * This software is a free software and there is NO WARRANTY.
/ * No restriction on use. You can use, modify and redistribute it for
/   personal, non-profit or commercial products UNDER YOUR RESPONSIBILITY.
/ * Redistributions of source code must retain the above copyright notice.
/
/-------------------------------------------------------------------------*/


#include "cpu.h"
#include "diskio.h"
#include "monitor.h"

#define	_USE_CD		1	/* J[hoXCb`gp */
#define _USE_WP		0	/* CgveNgXCb`gp */


/* MMC/SDR}h */
#define CMD0	(0)			/* GO_IDLE_STATE */
#define CMD1	(1)			/* SEND_OP_COND (MMC) */
#define	ACMD41	(0x80+41)	/* SEND_OP_COND (SDC) */
#define CMD8	(8)			/* SEND_IF_COND */
#define CMD9	(9)			/* SEND_CSD */
#define CMD10	(10)		/* SEND_CID */
#define CMD12	(12)		/* STOP_TRANSMISSION */
#define ACMD13	(0x80+13)	/* SD_STATUS (SDC) */
#define CMD16	(16)		/* SET_BLOCKLEN */
#define CMD17	(17)		/* READ_SINGLE_BLOCK */
#define CMD18	(18)		/* READ_MULTIPLE_BLOCK */
#define CMD23	(23)		/* SET_BLOCK_COUNT (MMC) */
#define	ACMD23	(0x80+23)	/* SET_WR_BLK_ERASE_COUNT (SDC) */
#define CMD24	(24)		/* WRITE_BLOCK */
#define CMD25	(25)		/* WRITE_MULTIPLE_BLOCK */
#define CMD55	(55)		/* APP_CMD */
#define CMD58	(58)		/* READ_OCR */


/* MMC|[gE@\ݒ  (vbgtH[ˑ`) */

#define CS_LOW()	*PGDR1 &= ~0x0010	/* MMC CS = L */
#define	CS_HIGH()	*PGDR1 |=  0x0010	/* MMC CS = H */
#define SOCKPORT	*PJPR0			/* \Pbgԓ(PJ) */
#define SOCKWP		0x0008			/* CgveNgړ_ (PJ3) */
#define SOCKINS		0x0002			/* J[hoړ_ (PJ1) */

#define BCLK		48000000UL		/* SPIW[NbN */
#define CLK_FAST	24000000UL		/* SCLK */
#define	CLK_SLOW	400000UL		/* SCLKᑬ () */
#define	DIV_F		(BCLK/CLK_FAST/2-1)
#define	DIV_S		(BCLK/CLK_SLOW/2-1)
#define	FCLK_SLOW()	{*SPCR0&=~0x40; *SPBR0=DIV_S; *SPCR0|=0x40;}	/* Set slow clock (100k-400k) */
#define	FCLK_FAST()	{*SPCR0&=~0x40; *SPBR0=DIV_F; *SPCR0|=0x40;}	/* Set fast clock (MMC:20MHz max, SD:25MHz max) */


/*--------------------------------------------------------------------------

   Module Private Functions

---------------------------------------------------------------------------*/

static volatile
DSTATUS Stat = STA_NOINIT;	/* hCuXe[^X */

static volatile
BYTE Timer1, Timer2;	/* 100Hz^C} (disk_timerproc()) */

static
BYTE CardType;			/* J[h^Cv */



/*-----------------------------------------------------------------------*/
/* MMCփf[^𑗐M  (vbgtH[ˑ֐)                         */
/*-----------------------------------------------------------------------*/

/* 1oCgM */
static
void xmit_spi (
	BYTE dat	/* Mf[^ */
)
{
	*(volatile BYTE*)SPDR0 = dat;	/* MJn */
	while ((*SPSR0 & 0x80) == 0) ;	/* ]I҂ */
	*(volatile BYTE*)SPDR0;		/* Mf[^͊Ă */
}

/* ubNM */
static
void xmit_spi_multi (
	const BYTE *buff,	/* Mf[^ւ̃|C^ */
	UINT btx			/* MoCg */
)
{
	*(volatile BYTE*)SPDR0 = *buff++;	/* MJn(pCvC1oCgl߂Ă) */
	btx--;

	do {
		*(volatile BYTE*)SPDR0 = *buff++;	/* ̃f[^𑗐M */
		while ((*SPSR0 & 0x80) == 0) ;		/* 1oCg]҂ */
		*(volatile BYTE*)SPDR0;			/* Mf[^Ă */
	} while (--btx);						/* f[^JԂ */

	while ((*SPSR0 & 0x80) == 0) ;			/* cMf[^Ă */
	*(volatile BYTE*)SPDR0;
}



/*-----------------------------------------------------------------------*/
/* MMCf[^M  (vbgtH[ˑ֐)                       */
/*-----------------------------------------------------------------------*/

/* 1oCgM */
static
BYTE rcvr_spi (void)
{
	*(volatile BYTE*)SPDR0 = 0xFF;	/* 0xFFM */
	while ((*SPSR0 & 0x80) == 0) ;		/* ]҂ */
	return *(volatile BYTE*)SPDR0;	/* Mf[^Ԃ */
}

/* ubNM */
static
void rcvr_spi_multi (
	BYTE *buff,		/* f[^Mobt@ւ̃|C^ */
	UINT btr		/* MoCg */
)
{
	*(volatile BYTE*)SPDR0 = 0xFF;	/* 0xFFM(pCvC1oCgl߂Ă) */
	btr--;

	do {
		*(volatile BYTE*)SPDR0 = 0xFF;	/* 0xFFM */
		while ((*SPSR0 & 0x80) == 0) ;		/* M҂ */
		*buff++ = *(volatile BYTE*)SPDR0;	/* Mf[^obt@ɃXgA */
	} while (--btr);						/* f[^JԂ */

	while ((*SPSR0 & 0x80) == 0) ;			/* cf[^obt@ɃXgA */
	*buff++ = *(volatile BYTE*)SPDR0;
}




/*-----------------------------------------------------------------------*/
/* J[hReady҂                                                     */
/*-----------------------------------------------------------------------*/

static
int wait_ready (void)	/* 1:Ready, 0:^CAEg */
{
	Timer2 = 50;	/* ^CAEg = 500ms */

	rcvr_spi();		/* 1oCgǂ(DO쓮x) */
	do {
		if (rcvr_spi() == 0xFF) return 1;	/* Ready */
		/* ̃[v͎Ԃv܂. }`^XNł͓KX rot_rdq()}܂. */
	} while (Timer2);	/* 0xFFM^CAEg܂Ń|[O */

	return 0;	/* ^CAEg */
}



/*-----------------------------------------------------------------------*/
/* J[h̃fBZNgSPIoX                                     */
/*-----------------------------------------------------------------------*/

static
void deselect (void)
{
	CS_HIGH();		/* CS = H */
	rcvr_spi();		/* 1oCg̃NbN */
}



/*-----------------------------------------------------------------------*/
/* J[hIReady҂                                             */
/*-----------------------------------------------------------------------*/

static
int select (void)	/* 1:OK, 0:^CAEg */
{
	CS_LOW();		/* CS = H */
	if (!wait_ready()) {	/* ^CAEg玸s */
		deselect();
		return 0;
	}
	return 1;
}



/*-----------------------------------------------------------------------*/
/* MMC/SDJ[hH̑ (vbgtH[ˑ֐)                 */
/*-----------------------------------------------------------------------*/

static
void power_on (void)	/* MMCHgp\ɂ */
{
	*STBCR5 &= ~0x02;	/* SPI0W[L */

	/* SPI0W[I/Osɐڑ */
	*PFCR3 = 0x0003;						/* MISO0 */
	*PFCR2 = (*PFCR2 & 0x0F0F) | 0x3030;	/* RSPCK0,MOSI0 */
	*PGIOR1 |= 0x0010; *PGCR5 &= 0xFFF0;	/* MMC_CS */
	CS_HIGH();
	if (_USE_CD) {*PJIOR0 &= 0xFFFD; *PJCR0 &= 0xFF0F;}	/* MMC_CD */
	if (_USE_WP) {*PJIOR0 &= 0xFFF7; *PJCR0 &= 0x0FFF;}	/* MMC_WP */

	/* SPI0W[ */
	*SPPCR0 = 0;						/* AChlŒƃ[vobN[h: gp */
	*SPSCR0 = 0;						/* V[PX: gp */
	*SPDCR0 = 0x20;						/* FIFOANZX: oCg */
	*SPBFCR0 = 0xF0; *SPBFCR0 = 0x30;	/* FIFO */
	*SPCMD00 = 0x0700;					/* SPI[h: MODE0, 8bit, MSB first */
	*SPCR0 = 0x48;						/* SPICl[u(}X^[h) */
}


static
void power_off (void)	/* MMCH̎gpI */
{
	select();			/* J[h̓I҂ */
	deselect();

	*SPCR0 = 0;			/* SPI0W[̃fBZ[u */
	*STBCR5 |= 0x02;		/* SPI0W[ */
}



/*-----------------------------------------------------------------------*/
/* MMCf[^pPbgM                                           */
/*-----------------------------------------------------------------------*/

static
int rcvr_datablock (	/* 1:OK, 0:G[ */
	BYTE *buff,			/* Mf[^XgAobt@ */
	UINT btr			/* f[^ubN */
)
{
	BYTE token;


	Timer1 = 20;
	do {							/* DataStartg[N҂ (^CAEg=200ms) */
		token = rcvr_spi();
		/* ̃[v͎Ԃv܂. }`^XNł͓KX rot_rdq()}܂. */
	} while ((token == 0xFF) && Timer1);
	if(token != 0xFE) return 0;		/* DataStartg[N̂Ƃ͎s */

	rcvr_spi_multi(buff, btr);		/* f[^obt@ɃXgA */
	rcvr_spi(); rcvr_spi();			/* CRCĂ */

	return 1;						/* MOK */
}



/*-----------------------------------------------------------------------*/
/* f[^pPbgMMCɑM                                             */
/*-----------------------------------------------------------------------*/

#if _READONLY == 0
static
int xmit_datablock (	/* 1:OK, 0:s */
	const BYTE *buff,	/* M512oCg̃f[^ւ̃|C^ */
	BYTE token			/* gpg[N */
)
{
	BYTE resp;


	if (!wait_ready()) return 0;		/* Ready҂(^CAEg玸s) */

	xmit_spi(token);					/* g[N𑗐M */
	if (token != 0xFD) {				/* StopTrang[NȊÔƂ̓f[^𑗐M */
		xmit_spi_multi(buff, 512);		/* f[^ */
		xmit_spi(0xFF); xmit_spi(0xFF);	/* _~[CRC */

		resp = rcvr_spi();				/* f[^X|XM */
		if ((resp & 0x1F) != 0x05)		/* 󂯕tȂG[ */
			return 0;
	}
	return 1;
}
#endif /* _READONLY */



/*-----------------------------------------------------------------------*/
/* Send a command packet to MMC                                          */
/*-----------------------------------------------------------------------*/

static
BYTE send_cmd (		/* ߂l: R1X|X (bit7==1:Ms) */
	BYTE cmd,		/* R}hCfbNX */
	DWORD arg		/*  */
)
{
	BYTE n, res;


	if (cmd & 0x80) {	/* ACMD<n>̂Ƃ͐CMD55𑗂 */
		cmd &= 0x7F;
		res = send_cmd(CMD55, 0);
		if (res > 1) return res;
	}

	/* J[hI */
	deselect();
	if (!select()) return 0xFF;

	/* Send command packet */
	xmit_spi(0x40 | cmd);				/* R}hCfbNX */
	xmit_spi((BYTE)(arg >> 24));		/* [31..24] */
	xmit_spi((BYTE)(arg >> 16));		/* [23..16] */
	xmit_spi((BYTE)(arg >> 8));			/* [15..8] */
	xmit_spi((BYTE)arg);				/* [7..0] */
	n = 0x01;							/* _~[CRC + Stop */
	if (cmd == CMD0) n = 0x95;			/* LCRC (CMD0(0)) */
	if (cmd == CMD8) n = 0x87;			/* LCRC (CMD8(0x1AA)) */
	xmit_spi(n);

	/* R}hX|XM */
	if (cmd == CMD12) rcvr_spi();		/* CMD12̂Ƃ͒1oCgĂ */
	n = 10;								/* X|X҂(ő10oCg) */
	do
		res = rcvr_spi();
	while ((res & 0x80) && --n);

	return res;							/* ꂽX|XԂ */
}



/*--------------------------------------------------------------------------

   Public Functions

---------------------------------------------------------------------------*/


/*-----------------------------------------------------------------------*/
/* fBXNhCȕ                                              */
/*-----------------------------------------------------------------------*/

DSTATUS disk_initialize (
	BYTE drv		/* hCuԍ (0) */
)
{
	BYTE n, cmd, ty, ocr[4];


	if (drv) return STA_NOINIT;			/* VOhCû݃T|[g */
	if (Stat & STA_NODISK) return Stat;	/* \Pbg? */

	power_on();							/* MMCH̏ */
	FCLK_SLOW();						/* ᑬNbN */
	for (n = 10; n; n--) rcvr_spi();	/* _~[NbNM */

	ty = 0;
	if (send_cmd(CMD0, 0) == 1) {			/* J[hSPI[h/IdleԂɈڍs */
		Timer1 = 100;						/* ^CAEg:1b */
		if (send_cmd(CMD8, 0x1AA) == 1) {	/* SDv2? */
			/* ԂSDv2 */
			for (n = 0; n < 4; n++) ocr[n] = rcvr_spi();		/* R7 resp̖߂lM */
			if (ocr[2] == 0x01 && ocr[3] == 0xAA) {				/* J[h2.7-3.6Vœ\? */
				while (Timer1 && send_cmd(ACMD41, 1UL << 30)) ;	/* I҂(ACMD41(HCS)) */
				if (Timer1 && send_cmd(CMD58, 0) == 0) {		/* OCRǂCCSrbg`FbN */
					for (n = 0; n < 4; n++) ocr[n] = rcvr_spi();
					ty = (ocr[0] & 0x40) ? CT_SD2 | CT_BLOCK : CT_SD2;	/* SDv2AhbVO[h */
				}
			}
		} else {
			/* SDv2ȊO */
			if (send_cmd(ACMD41, 0) <= 1) 	{	/* SDv1MMC? */
				ty = CT_SD1; cmd = ACMD41;	/* SDv1 (ACMD41(0)) */
			} else {
				ty = CT_MMC; cmd = CMD1;	/* MMCv3 (CMD1(0)) */
			}
			while (Timer1 && send_cmd(cmd, 0)) ;		/* I҂ */
			if (!Timer1 || send_cmd(CMD16, 512) != 0)	/* Ô߃ubN512ɐݒ */
				ty = 0;
		}
	}
	CardType = ty;	/* J[h^CvtO */
	deselect();

	if (ty) {			/*  */
		FCLK_FAST();			/* NbN */
		Stat &= ~STA_NOINIT;	/* Xe[^XSTA_NOINITNA */
	} else {			/* s */
		power_off();
		Stat = STA_NOINIT;
	}

	return Stat;
}



/*-----------------------------------------------------------------------*/
/* fBXNXe[^X̎擾                                              */
/*-----------------------------------------------------------------------*/

DSTATUS disk_status (
	BYTE drv		/* hCuԍ (0) */
)
{
	if (drv) return STA_NOINIT;		/* VOhCû݃T|[g */

	return Stat;	/* fBXNXe[^XԂ */
}



/*-----------------------------------------------------------------------*/
/* f[^̓ǂݏo                                                      */
/*-----------------------------------------------------------------------*/

DRESULT disk_read (
	BYTE drv,		/* hCuԍ (0) */
	BYTE *buff,		/* f[^obt@ւ̃|C^ */
	DWORD sector,	/* ǂݏoJnZN^ԍ (LBA) */
	BYTE count		/* ǂݏoZN^ (1..255) */
)
{
	if (drv || !count) return RES_PARERR;		/* p[^L`FbN */
	if (Stat & STA_NOINIT) return RES_NOTRDY;	/* hCu\`FbN */

	if (!(CardType & CT_BLOCK)) sector *= 512;	/* AhbVO[hJ[hɍ킹 */

	if (count == 1) {	/* VOubNǂݏo */
		if ((send_cmd(CMD17, sector) == 0)	/* READ_SINGLE_BLOCK */
			&& rcvr_datablock(buff, 512))
			count = 0;
	}
	else {				/* }`ubNǂݏo */
		if (send_cmd(CMD18, sector) == 0) {	/* READ_MULTIPLE_BLOCK */
			do {
				if (!rcvr_datablock(buff, 512)) break;
				buff += 512;
			} while (--count);
			send_cmd(CMD12, 0);				/* STOP_TRANSMISSION */
		}
	}
	deselect();

	return count ? RES_ERROR : RES_OK;	/* ʂԂ */
}



/*-----------------------------------------------------------------------*/
/* f[^̏                                                      */
/*-----------------------------------------------------------------------*/

#if _READONLY == 0
DRESULT disk_write (
	BYTE drv,			/* hCuԍ (0) */
	const BYTE *buff,	/* ރf[^ւ̃|C^ */
	DWORD sector,		/* ݊JnZN^ԍ (LBA) */
	BYTE count			/* ރZN^ (1..255) */
)
{
	if (drv || !count) return RES_PARERR;		/* p[^L`FbN */
	if (Stat & STA_NOINIT) return RES_NOTRDY;	/* hCu\`FbN */
	if (Stat & STA_PROTECT) return RES_WRPRT;	/* hCu݉\`FbN */

	if (!(CardType & CT_BLOCK)) sector *= 512;	/* AhbVO[hJ[hɍ킹 */

	if (count == 1) {	/* VOubN */
		if ((send_cmd(CMD24, sector) == 0)	/* WRITE_BLOCK */
			&& xmit_datablock(buff, 0xFE))
			count = 0;
	}
	else {				/* }`ubN */
		if (CardType & CT_SDC) send_cmd(ACMD23, count);
		if (send_cmd(CMD25, sector) == 0) {	/* WRITE_MULTIPLE_BLOCK */
			do {
				if (!xmit_datablock(buff, 0xFC)) break;
				buff += 512;
			} while (--count);
			if (!xmit_datablock(0, 0xFD))	/* STOP_TRAN token */
				count = 1;
		}
	}
	deselect();

	return count ? RES_ERROR : RES_OK;	/* ʂԂ */
}
#endif /* _READONLY == 0 */



/*-----------------------------------------------------------------------*/
/* ̑̃hCu                                                  */
/*-----------------------------------------------------------------------*/

#if _USE_IOCTL != 0
DRESULT disk_ioctl (
	BYTE drv,		/* hCuԍ (0) */
	BYTE ctrl,		/* R[h */
	void *buff		/* f[^ւ̃|C^ */
)
{
	DRESULT res;
	BYTE n, csd[16], *ptr = buff;
	WORD csize;


	if (drv) return RES_PARERR;					/* p[^L`FbN */
	if (Stat & STA_NOINIT) return RES_NOTRDY;	/* hCu\`FbN */

	res = RES_ERROR;

	switch (ctrl) {
	case CTRL_SYNC :		/* hCu̓҂ */
		if (select()) {
			deselect();
			res = RES_OK;
		}
		break;

	case GET_SECTOR_COUNT :	/* hCue(ZN^)Ԃ (DWORD) */
		if ((send_cmd(CMD9, 0) == 0) && rcvr_datablock(csd, 16)) {
			if ((csd[0] >> 6) == 1) {	/* SDC ver 2.00 */
				csize = csd[9] + ((WORD)csd[8] << 8) + 1;
				*(DWORD*)buff = (DWORD)csize << 10;
			} else {					/* SDC ver 1.XX or MMC ver 3 */
				n = (csd[5] & 15) + ((csd[10] & 128) >> 7) + ((csd[9] & 3) << 1) + 2;
				csize = (csd[8] >> 6) + ((WORD)csd[7] << 2) + ((WORD)(csd[6] & 3) << 10) + 1;
				*(DWORD*)buff = (DWORD)csize << (n - 9);
			}
			res = RES_OK;
		}
		break;

	case GET_SECTOR_SIZE :	/* ZN^TCY(oCg)Ԃ (WORD) */
		*(WORD*)buff = 512;
		res = RES_OK;
		break;

	case GET_BLOCK_SIZE :	/* ubNTCY(ZN^)Ԃ (DWORD) */
		if (CardType & CT_SD2) {	/* SDC ver 2.00 */
			if (send_cmd(ACMD13, 0) == 0) {	/* Read SD status */
				rcvr_spi();
				if (rcvr_datablock(csd, 16)) {				/* Read partial block */
					for (n = 64 - 16; n; n--) rcvr_spi();	/* Purge trailing data */
					*(DWORD*)buff = 16UL << (csd[10] >> 4);
					res = RES_OK;
				}
			}
		} else {					/* SDC ver 1.XX or MMC */
			if ((send_cmd(CMD9, 0) == 0) && rcvr_datablock(csd, 16)) {	/* Read CSD */
				if (CardType & CT_SD1) {	/* SDC ver 1.XX */
					*(DWORD*)buff = (((csd[10] & 63) << 1) + ((WORD)(csd[11] & 128) >> 7) + 1) << ((csd[13] >> 6) - 1);
				} else {					/* MMC */
					*(DWORD*)buff = ((WORD)((csd[10] & 124) >> 2) + 1) * (((csd[11] & 3) << 3) + ((csd[11] & 224) >> 5) + 1);
				}
				res = RES_OK;
			}
		}
		break;

	/* ȍ~ioctlR}hFatFsł͎gpȂ */

	case MMC_GET_TYPE :		/* J[h^CvԂ (BYTE) */
		*ptr = CardType;
		res = RES_OK;
		break;

	case MMC_GET_CSD :		/* CSDǂݏo (16oCg) */
		if (send_cmd(CMD9, 0) == 0		/* READ_CSD */
			&& rcvr_datablock(ptr, 16))
			res = RES_OK;
		break;

	case MMC_GET_CID :		/* CIDǂݏo (16oCg) */
		if (send_cmd(CMD10, 0) == 0		/* READ_CID */
			&& rcvr_datablock(ptr, 16))
			res = RES_OK;
		break;

	case MMC_GET_OCR :		/* OCRǂݏo (4oCg) */
		if (send_cmd(CMD58, 0) == 0) {	/* READ_OCR */
			for (n = 4; n; n--) *ptr++ = rcvr_spi();
			res = RES_OK;
		}
		break;

	case MMC_GET_SDSTAT :	/*SDXe[^Xǂݏo (64oCg) */
		if (send_cmd(ACMD13, 0) == 0) {	/* SD_STATUS */
			rcvr_spi();
			if (rcvr_datablock(ptr, 64))
				res = RES_OK;
		}
		break;

	default:
		res = RES_PARERR;
	}

	deselect();

	return res;
}
#endif /* _USE_IOCTL != 0 */


/*-----------------------------------------------------------------------*/
/* ^C}荞ݏ  (vbgtH[ˑ֐)                        */
/*-----------------------------------------------------------------------*/
/* ̊֐̓^C}荞ݏ̒疈b100ĂяoAJ[h
/  ^C~O𓾂J[ȟoȂǂsB
*/

void disk_timerproc (void)
{
	static WORD pv;
	BYTE n, s;


	n = Timer1;						/* 100HzZ^C}(0~) */
	if (n) Timer1 = --n;
	n = Timer2;
	if (n) Timer2 = --n;

	n = pv;
	pv = SOCKPORT & (SOCKWP | SOCKINS);	/* \PbgԂ̃TvO */

	if (n == pv) {					/* ړ_Ԃ͈肵? */
		s = Stat;
		if (_USE_WP && (pv & SOCKWP))	/* WP == H (݋֎~) */
			s |= STA_PROTECT;
		else							/* WP == L (݋) */
			s &= ~STA_PROTECT;

		if (_USE_CD && (pv & SOCKINS))	/* CD == H (J[hȂ) */
			s |= (STA_NODISK | STA_NOINIT);
		else							/* CD == L (J[h) */
			s &= ~STA_NODISK;

		Stat = s;
	}
}

